;-------------------------------------------------------------
; RAM-Floppy-Treiber 
; RAF der Akademie der Wissenschaften, 
; Speicherkapazitaet von 128, 512 o. 2048 KByte
;-------------------------------------------------------------
; Treiber reassembliert V. Pohlers 2009
; Treiber RAFCPMP (mit Paritaet)
;-------------------------------------------------------------
;
; GIDE-Treiber nach Original GIDE-Quellen neu gebaut
; H. Poppe, 2014 / 2015
; Funktionen: 
; Erkennung freier LW-Buchstaben im System
; automatische Zuordnung der GIDE-Laufwerke im CP/M
; Erkennung des angeschlossenen IDE-Laufwerkes
;
; Stand: 31.12.2014
; 
;  
;-------------------------------------------------------------
; assemblieren:
; 	m80.com =treiber
; 	link.com treiber[op]
; 	prl2com treiber.prl
;  oder mit dem make.cmd von V.P.
;-------------------------------------------------------------
	;
;	.Z80
	cseg
	;
;GIDE	equ	060h		; GIDE Basisadresse
	;
aktlw	equ	04h		;Merkzelle Laufwerk + User
bdos	equ	05h		;BDOS-Aufruf
	;
	; BDOS-Aufrufe
	;
conin	equ	01h		;Konsoleneingabe
conout	equ	02h		;Konsolenausgabe
zkout	equ 	09h		;Zeichenkette ausgeben
vernr	equ 	0Ch		;Abfrage CP/M-Version
	;
	; BS-Steuerzeichen
	;
cr	equ	0Dh
lf	equ	0Ah	
	;
dma	equ	80h		;Standard-DMA
	;
ndisk	equ 16		;maximale Anzahl Laufwerke
	;
	; Definition der GIDE I/O Addressen:
	;
IdeDOR	equ	GIDE+6		; Digital Output Register
IdeDat	equ	GIDE+8		; Data Register
IdeErr	equ	GIDE+9		; Error Register
IdeWrPc	equ	IdeErr		; Write Precomp Register
IdeSCnt	equ	GIDE+lf	; Sector Count
IdeSNum	equ	GIDE+0Bh	; Sector Number
IdeCLo	equ	GIDE+0Ch	; Cylinder Low
IdeCHi	equ	GIDE+cr	; Cylinder High
IdeSDH	equ	GIDE+0Eh	; Drive and thead
IdeCmd	equ	GIDE+0Fh	; Command
IdeStat	equ	GIDE+0Fh	; Status
	;
	; Kommandos fuer IDE Laufwerke:
	;
CmdRd	equ	20h		; Read Sector
CmdWr	equ	30h		; Write Sector
CmdInit	equ	91h		; Initialize Drive Params
	;
	;
	;bdos constants on entry to write (passed in C)
	;
wrall	equ	0		;write to allocated
wrdir	equ	1		;write to directory
wrual	equ	2		;write to unallocated
	;
	;-----------------------------------------------------
	; Initialisierung
	;-----------------------------------------------------

init:	pop	hl		;Return-Adresse vom CCP
	ld	(hwboo1+1), hl	;Warmstart NEU
	ld	(hwboot+1), sp	
	push	hl
	ld 	sp,init		;SP vor Programmbeginn
	;
	ld	de, tdrv 	; Nachladbarer ...
	ld	c,zkout
	call	bdos
	;
	; Version abfragen
	;
	ld c,vernr		;CP/M Version
	call bdos
	ld (versn),hl		;H=0, L=2x oder 3x
	ld a,l
	and 0F0h
	cp 20h
	jr z,init1
;	cp 30h
;	jr z,init1
	ld de,tnver		; falsche Version
	jp init4		; Ende
	;
init1:	call	dcheck		; freie Laufwerke suchen
	;
	call gaddr		; GIDE Adresse
	call hdinfo		;IDE-Erkennung
	jr nc,init2
	;
	jp 0
	;
init2:	call	dvzeig		; vorhandene LW anzeigen
	call	hdset		; HDD einbinden
	call	dnzeig		; neue LW anzeigen
	;
	;
	;-----------------------------------------------------
	; Kommandozeilenparameter lesen (wozu?)
	;
	ld	hl,dma	; Kommandozeile analysieren
	ld	a,(hl)
	ld	b,a		; Anzahl
	or	a
	jr	z,init3		; Z = keine Parameter
zpar1:	inc	hl
	ld	a,(hl)		; 1. Zeichen testen
	cp	' '		; Leerzeichen ?
	jr	z,zpar2
	cp	'/'
	jr	z,zpar2	; / ?
	jr	init3		; Zeichen erkannt
zpar2:	djnz	zpar1		; Leerz. am Anfang uebergehen
	;
	;-----------------------------------------------------
	;
init3:	jr	init10
	;
init4:	ld	c,zkout
	call	bdos
	jp	0		; Warmstart (auch nach INIT)
	;
init10:	nop
	;
	;BIOS patchen
	;
	di
	ld	hl, (1)		; Warmstartvektor
	dec	hl
	dec	hl
	dec	hl
	push	hl			; Adresse BOOT (Beginn BIOS)
; BIOS-Einspruenge sichern
	ld	de, biosv	
	ld	bc, 33h
	ldir
; BIOS-Funktionen patchen
	pop	ix			; Adresse BOOT (Beginn BIOS)
	di
	ld	de, hwboot
	ld	(ix+4),	e
	ld	(ix+5),	d
	ld	de, hhome
	ld	(ix+19h), e
	ld	(ix+1Ah), d
	ld	de, hsldsk
	ld	(ix+1Ch), e
	ld	(ix+1Dh), d
	ld	de, hsetrk
	ld	(ix+1Fh), e
	ld	(ix+20h), d
	ld	de, hsetsc
	ld	(ix+22h), e
	ld	(ix+23h), d
	ld	de, hsetdm
	ld	(ix+25h), e
	ld	(ix+26h), d
	ld	de, hread
	ld	(ix+28h), e
	ld	(ix+29h), d
	ld	de, hwrite
	ld	(ix+2Bh), e
	ld	(ix+2Ch), d
	ld	de, hsctrn
	ld	(ix+31h), e
	ld	(ix+32h), d
	ei
	;
	ld	hl, (6)		; orig. BDOS-Einsprung
	ld	(rbdos+1), hl	; in RBDOS patchen
	ld	a, l
	cp	6		; bdos steht auf Adr xx06h
	jr	nz,init12
	inc	hl
	inc	hl
	inc	hl
	push	hl
	ld	de,tab1
	ld	bc,8
	ldir
	ld	hl,tab2
	pop	de
	ld	bc,8
	ldir
init12:	ld	hl, rbdos
	ld	(6), hl
	ld	de, tdlw 		; GIDE installiert
	jp	init4			; ENDE
	;
	;-----------------------------------------------------
	;
	; Test auf vorhandene Laufwerke
	;
	; BIOS-Adresse fuer SELDSK berechnen (fuer CPM 2.2)
	;
dcheck:	ld c,19h		;aktuelles Laufwerk
	call bdos
	ld (lwalt),a		;merken
	ld hl,(01h)		;Warmstart-Adresse
	ld de,24		;Abstand zu SELDSK
	add hl,de
	ld (seldsk),hl		;Adresse merken
	;
	ld hl,lwtab		;LW-Tabelle
	ld b,16
dche01:	ld (hl),0FFh		;Tabelle loeschen
	inc hl
	djnz dche01
	ld hl,lwtab
	ld b,16			;max. Anzahl phys. LW
	ld c,0			;erstes LW A: = 0
dche02:	push hl
	push bc
	ld hl,(versn)
	ld a,l
	and 0F0h
	cp 20h
	jr nz,dche03
	call dchv2		;Version 2.x
	jr dche04
dche03:	cp 30h
	jr nz,dche06
	call dchv3		;Version 3.x
dche04:	ld a,h
	or l
	pop bc
	pop hl
	jr z,dche06		;Z = kein LW installiert
	ld (hl),c		;in Tabelle schreiben
dche06:	inc hl
	inc c
	djnz dche02
	;
	ld c,cr		;LW ruecksetzen
	call bdos
	;
	ld a,(lwalt)
	ld e,a
	ld c,0Eh		;LW setzen
	call bdos
	;
	ret
	;
	; SELDSK fuer CP/M 2.2
	;
dchv2:	ld hl,dchret		;Return-Adresse
	push hl
	ld hl,(seldsk)		;BIOS-Adresse
	jp (hl)			;Aufruf BIOS
	;
dchret:	nop
	ret
	;
	; SELDSK fuer CP/M 3.x
	;
dchv3:	ld a,9			;BIOS Funktionsnummer
	ld (biospb),a
	ld a,1
	ld (brgde),a		;E =1
	ld a,c
	ld (brgbc),a		;LW-Nr.
	ld de,biospb
	ld c,32h		;Direct BIOS Call
	call bdos
	ret
	;
	;		;Parametertabelle
biospb:	defb 0		;BIOS Funktionsnummer
	defb 0		;Inhalt A
brgbc:	defw 0		;Inhalt BC
brgde:	defw 0		;Inhalt DE
	defw 0		;Inhalt HL
versn:	defw 0		;CP/M-Versionsnummer
seldsk:	defw 0		;BIOS-Adresse fuer LW-Select
lwalt:	defb 0		;aktuelles Laufwerk
	;
ms:	defb 0
errst:	defb 0	
	;
	;-----------------------------------------------------
	;
	; neue HDD in LW-Tabelle einbinden
	;
hdset:	ld hl,lwtab	;
	inc hl
	inc hl		;erst ab C: beginnen
	ld e,0A0h	;erste HDD-Kennung
	ld b,14		;max. LW -2
	ld a,(hdtab)	;Anz. HDD-LW
	ld c,a
hdse1:	ld a,(hl)
	cp 0FFh		;freier LW-Buchst.?
	jr nz,hdse2	;NZ = nein
	ld (hl),e	;HDD einsetzen
	inc e		;naechste HDD
	dec c		;Zaehler HDD-1
hdse2:	inc hl
	ret z
	djnz hdse1
	ret
	;
	; Tabelle der Festplattenlaufwerke
	;
hdtab:	defb	8	;Anzahl HDD-Laufwerke
	;
	; installierte Laufwerke anzeigen
	;
dvzeig:	ld de,tvlw
	ld c,zkout
	call bdos
	ld hl,lwtab
	ld b,16			;max. Anz. LW
dvzei1:	ld a,(hl)
	cp 0FFh
	jr z,dvzei2		;FF = kein LW
	push bc
	push hl
	ld e,41h		;LW-Nr in Buchstabe
	add a,e
	call dazeig
	pop hl
	pop bc
dvzei2:	inc hl
	djnz dvzei1
	jr dazei9
	;
	; neue Laufwerke anzeigen
	;
dnzeig:	ld de,tnlw
	ld c,zkout
	call bdos
	ld hl,lwtab
	ld b,16
	ld c,0			;Zaehler fuer LW-Buchst.
dnzei1:	ld a,(hl)
	and 0F0h
	cp 0A0h
	jr nz,dnzei2		;NZ -> kein neues LW
	push bc
	push hl
	ld a,41h
	add a,c
	call dazeig
	pop hl
	pop bc
dnzei2:	inc hl
	inc c
	djnz dnzei1
	jr dazei9
	;
dazeig:	ld e,a
	ld c,conout
	call bdos
	ld de,tdpsp
	ld c,zkout
	call bdos
	ret
	;
dazei9:	ld de,tcrlf
	ld c,zkout
	call bdos
	ret
	;
	; GIDE Adresse zeigen
	;
gaddr:	ld de,taddr	;GIDE Adresse ...
	ld c,zkout
	call bdos
	ld a,GIDE		;Basisadresse
	call outa		;ausgeben
	ld a,'H'
	call ochr
	ld de,tcrlf
	ld e,zkout
	call bdos
	ret
	;
	; HDINFO - Erkennung des ID-Laufwerkes
	;
hdinfo:	ld a,0A0h	;Master-Kennung
	ld (ms),a	;merken
	ld a,6
	out (idedor),a	;Software-Reset
	ld b,0
hdin01:	djnz hdin01	;warten
	in a,(idecmd)	;Status abfragen
hdin02:	djnz hdin02
	ld a,2
	out (idedor),a	;Reset aufheben
	ld hl,4000h
hdin03:	djnz hdin03
	dec hl
	ld a,h
	or l
	jp z,nordy	;Abbruch
	in a,(idecmd)
	and 11010001b
	cp 01010000b	;Test
	jr nz,hdin03
	ld a,90h	;Diagnose
	out (idecmd),a
	call wait
	jp c,nordy
	in a,(ideerr)	;Ergebnis
	ld (errst),a	;merken
	ld a,(errst)
	and 07Fh
	cp 01h
	jp nz,nordy
	;
	ld a,(ms)	;Master/Slave-Kennung
	out (idesdh),a	;ausgeben
	ld a,0ECh	;Befehl Laufwerksinfo lesen
	out (idecmd),a
	call wait
	jp c,nordy
	;
	ld hl,hdbuf	;Lesepuffer
	ld b,0
	ld c,idedat	;Portadresse
	inir
	inir		;512 Byte einlesen
	;
	;Laufwerksinfo auf 0 testen
	;
	ld hl,hdbuf
	ld a,(hl)	;Konf-Byte 0
	or a
	jp z,nordy
	inc hl
	;		;Konf-Byte 1 nicht testen
	inc hl
	ld a,(hl)
	inc hl
	or (hl)
	jp z,nordy	;Zylinder=0
	inc hl
	inc hl
	inc hl
	ld a,(hl)
	inc hl
	or (hl)
	jp z,nordy	;Kopfanzahl=0
	inc hl
	inc hl
	inc hl
	inc hl
	inc hl
	ld a,(hl)
	inc hl
	or (hl)
	jp z,nordy	;Sektoranzahl=0
	ld hl,hdbuf+54
	ld b,40		;Modellnummer, 20 Words
test7:	bit 7,(hl)
	jp nz,nordy	;Bit 7 in Modellnummer
	inc hl
	djnz test7
	;
	;Laufwerksinfo
	;
	ld de,tlide	;IDE-Laufwerk
	ld c,zkout
	call bdos
	ld de,tspc2
	ld c,zkout
	call bdos
	;
	ld hl,hdbuf+54	;String-Anfang
	ld de,dirbf	;DEBUG
	ld bc,128	;umladen
	ldir
	;
	ld hl,hdbuf+54	;String-Anfang
	ld b,24		;String-Laenge
exwd:	nop
	ld a,(hl)
	call check
	inc hl
	ld c,(hl)	;Byte-Reihenfolge tauschen
	ld (hl),a
	dec hl
	ld a,c
	call check
	ld (hl),a
	inc hl
	inc hl
	djnz exwd
	ld hl,hdbuf+94
	ld (hl),'$'	;String-Ende setzen
	;
	; Modell anzeigen
	;
	ld de,hdbuf+54	;String-Anfang
	ld c,zkout
	call bdos
	ld de,tcrlf
	ld c,zkout
	call bdos
	;
	;Plattengeometrie anzeigen
	;
	ld de,tspc2
	ld c,zkout
	call bdos
	ld hl,(hdbuf+2)		;Zylinder
	call hldez
	ld de,tcyl
	ld c,zkout
	call bdos
	ld hl,(hdbuf+6)		;Koepfe
	call hldez
	ld de,thead
	ld c,zkout
	call bdos
	ld hl,(hdbuf+12)	;Sektoren
	call hldez
	ld de,tsecs
	ld c,zkout
	call bdos
	;
	ret
	;
	; vereinfachte Fehlerbehandlung
	;
nordy:	ld de,tfehl
	ld c,zkout
	call bdos
	scf
	ret
	;
	;warten auf Platte
	;
wait:	ld de,3000h	;ca 10 sek
	ld b,0
wait4:	djnz wait4
	dec de
	ld a,d
	or e
	jr z,timout
	in a,(idecmd)
	rla
	jr c,wait4
	ret
	;
timout:	scf
	ret
	;
	; Test auf ungueltige Zeichen
	; und Austausch durch '?'
	;
check:	or a
	ret z
	cp 20h
	jr nc,check2
check1:	ld a,'?'
check2:	cp 7Fh
	jr nc,check1
	ret
	;
	;
	; Dezimalausgabe des Wertes in HL
	; PE:    HL   Zahl
	;
HLDEZ:	PUSH DE             ; bei Rekursion: letzten Rest retten
	LD   BC,-10
	LD   DE,-1          ; Zaehler fuer Ergebnis (Anzahl Sub'
HLDEZ1: ADD  HL,BC
	INC  DE
	JR   C,HLDEZ1       ; DE = HL/10
	LD   BC,10
	ADD  HL,BC          ; letzten Abzug korrigieren; HL = Re
	EX   DE,HL          ; 10er / Rest tauschen
	LD   A,H
	OR   L              ; Zehner = 0 ?
	CALL NZ,HLDEZ       ; Rekursion wenn Zehner vorhanden
	LD   A,E
	ADD  A,'0'          ; Rest -> ASCII in A
	call ochr
	POP  DE             ; voriger Rest
	RET
		;
	; Ausgabe von A als HEX-Wert
	;
outa:	push af
	RRA
	RRA
	RRA
	RRA
	CALL AHEX0
	POP  AF
AHEX0:	AND  0Fh
	add  a,90h
	daa
	adc  a,40h
	daa
	call ochr
	ret
	;
ochr:	push af
	push bc
	push de
	push hl
	ld e,a
	ld c,conout
	call bdos
	pop hl
	pop de
	pop bc
	pop af
	ret
	;
;-------------------------------------------------------------
;	; Texte
;-------------------------------------------------------------
	;
tdrv:	defm cr,lf,'GIDE-Treiber fuer CP/M 2.x  V 0.2',cr,lf,'$'
tdlw:	defm cr,lf,'GIDE-Treiber installiert',cr,lf,'$'
tvlw:	defm cr,lf,'Im BIOS installierte Laufwerke:',cr,lf,'$'
tnlw:	defm cr,lf,'Neu installierte GIDE-Laufwerke:',cr,lf,'$'
tnver:	defm cr,lf,'falsche CP/M-Version ',cr,lf,'$'
tdpsp:	defm ': $'
tcrlf:	defm cr,lf,'$'
tspc2:	defm '  $'
	;
	; Texte fuer IDE-Erkennung
	;
taddr:	defm cr,lf,'  GIDE-Modul auf Adresse $'
tlide:	defm cr,lf,'  Info IDE-Laufwerk: ',cr,lf,'$'
tcyl:	defm ' Cyl; $'
thead:	defm ' Head; $'
tsecs:	defm ' Sec;',cr,lf,'$'
	;
tfehl:	defm cr,lf,'Fehler bei der Laufwerkserkennung $'
	;
	;-----------------------------------------------------
	;
jphl:	jp	(hl)
	;
tab2:	dw	loc_611
	dw	loc_615
	dw	loc_619
	dw	loc_61D
	;
	defm 'Ende Vorspann'
	;
	;-----------------------------------------------------
	; BDOS
	;-----------------------------------------------------
	;	align 100h
	ORG	($+255)/256*256
	;
	db	0FEh
	db	0FEh
	db	0FEh
	db	0FEh
	db	0FEh
	db	0FEh
	;
	; hier Adr xxx06, wird in init getestet
	;
rbdos:	jp	5

tab1:	dw	0
	dw	0
	dw	0
	dw	0
;
loc_611:	ld	hl, (tab1)
		jp	(hl)
loc_615:	ld	hl, (tab1+2)
		jp	(hl)
loc_619:	ld	hl, (tab1+4)
		jp	(hl)
loc_61D:	ld	hl, (tab1+6)
		jp	(hl)

;-------------------------------------------------------------
;BIOS-Sprungverteiler:
;wird mit dem originalen Sprungverteiler des BIOS gefuellt
;dann werden im originalen Sprungverteiler des BIOS die 
;GIDE-spezifischen Einspruenge gepatcht
;-------------------------------------------------------------
;
biosv:
	defs	3		;cold boot
	defs	3		;warm boot
	defs	3		;console status
	defs	3		;console input
	defs	3		;console output
	defs	3		;list output
	defs	3		;punch output
	defs	3		;reader input
fhome:	defs	3		;home disk drive
fsldsk:	defs	3		;select drive
fsttrk:	defs	3		;set track
fstsec:	defs	3		;set sector
fstdma:	defs	3		;set dma address
fread:	defs	3		;read sector
fwrite:	defs	3		;write sector
flists:	defs	3		;list status
fsctrn:	defs	3		;sector translate
;
;-------------------------------------------------------------
;BIOS-Routinen
;-------------------------------------------------------------
	;
	; BOOT
hwboot:	ld	sp,0		;Stack: wird gepatcht
	ld	hl, rbdos	;BDOS Einsprung NEU
	ld	(6), hl
	;
	ld	c,0Dh		;DISK RESET
	call	5
	;
	; WBOOT
	ld hl,0000		;0000 in HL
	ld (unacnt),hl		;clr unalloc cnt, 
	;			;hst buf inactive
hwboo1:	jp	0		;zurueck zum CCP
	;
newstk:	defw	0		;Merkzelle Stack
	;
;*************************************************************
;Some fixed data options.
;
;lwtab -- array of select values for mapping the disk number
;	passed to SELDSK to either the floppy-disk bios or the
;	hard disk bios--selecting which disk to use.
;
;	00-0F = Disk im BIOS installiert
;	A0-AF = Master IDE drive
;	B0-BF = Slave IDE drive
;	FF    = Laufwerk nicht installiert
;	Initialisierung mit FF, wird erst
;	beim Laden des Treibers gefuellt
;
	defm ' Laufwerktabelle '
;
lwtab:	defs	16	
;
;*************************************************************
;Disk-Routinen fuer GIDE *** NEU ***
;
;*************************************************************
;Function:  seldsk
;Purpose:   Select disk drive
;Entry:     Disk number 0-F in C
;Exit:      Address of DPH in HL or 0000 if error
;Used:      All
;
hsldsk:	ld a,c			;LW-Nummer
	cp ndisk		;Vergleich mit max. LW-Anz.
	jr nc,errsel		;Fehler, wenn Wert >16
	call mkesel		;holt Wert aus Disk-Tabelle
	inc a			;Test auf 0FFh
	jr z,errsel		;wenn Z=1, keine Disk
	dec a			;Wert wiederherstellen
	push af
	ld a,(aktlw)		;altes LW+User-Nr.
	and 0F0h		;User-Nr. filtern
	or c			;neue LW-Nr. dazu
	ld (aktlw),a		;neues LW merken
	pop af
	rlca			;Test auf Bit 7
	rrca			;in Carry
	jr c,idesel		;Bit 7 = IDE-Disk
	;
	ld c,a			;floppy disk, put back in C
	xor a			;0 in A
	ld (hdisk),a		;wenn hdisk=0
	jp fsldsk		;dann ins BIOS
	;
idesel:	ld	(hdisk),a	;flag as our disk
	ld	h,0		;make sure H=0
	and	0Fh		;HD-Nummer maskieren
	ld	l,a		;and put in L
	ld	a,c		;get disk number again
	ld	(sekdsk),a	;save disk number
	add	hl,hl		;*2
	add	hl,hl		;*4
	add	hl,hl		;*8
	add	hl,hl		;*16
	ld	de,hdpbase	;get base of hard disk dph's
	add	hl,de		;make pointer to DPH for drive
	ret			;and go back
	;
errsel:	ld	hl,0000		;0 in HL
	ld	a,(aktlw)	;get current disk
	sub	c		;same as selected?
	ret	nz		;wasn't same
	ld	(aktlw),a	;else force drive A current
	ret			;and return with error
	;
mkesel:	ld	b,0		;0 in B
	ld	hl,lwtab	;base of select array in HL
	add	hl,bc		;make pointer to select data
	ld	a,(hl)		;get the select data
	ret
	;
;*************************************************************
;Function:  Home
;Purpose:   Home current disk
;Entry:     None
;Exit:      Disk homed
;Uses:      AF, HL
;
hhome:	call	chkhrd		;is it our disk?
	jp	z,fhome		;no, jump to Floppy
	xor 	a		;0 in A
	ld	hl,hstact	; HL points to host active
	or	(hl)		;set flags
	ret	z		;return if host is not active
	inc	hl		;bump to hstwrt
	xor	a		;0 in A
	or	(hl)		;set flagz
	ret	nz		;return if pending write
	dec	hl		;bump back down to hstact
	ld	(hl),a		; else clear host active flag
	ret			;and done
	;
;*************************************************************
;Function:  Hsettrk
;Purpose:   Set logical track number for disk I/O
;Entry:     BC = logical sector number
;Exit:      track set up
;Used:      All
;
Hsetrk:	call	chkhrd		;is it our disk?
	jp	z,Fsttrk	;no, jump to Fbios
	;
	ld	(sektrk),bc	;save it
	ret			;and done
	;
;*************************************************************
;Function:  Hsetsec
;Purpose:   Set logical sector number for disk I/O
;Entry:     BC = logical sector number
;Exit:      sector set up
;Used:      All
;
Hsetsc:	call	chkhrd		;is it our disk?
	jp	z,Fstsec	;not our disk.  Jump to FBIOS
	;
	ld	(SEKSEC),bc	;save sector to seek
	ret
	;
;*************************************************************
;Function:  Hsetdma
;Purpose:   Set transfer address for disk I/O
;Entry:     BC = dma address
;Exit:      DMA address set up
;Used:      All
;
Hsetdm:	ld	(dmaadr),bc	;set it here
	jp	Fstdma		;and let FBIOS set it as well.
;	
;*************************************************************
;Function:  Hsctrn
;Purpose:   Do logical->physical sector translation
;Entry:     BC = logical sector number, DE points to xlate table
;Exit:      HL = physical sector number
;Uses:      All
;
Hsctrn:	call	chkhrd		;is it our disk?
	jp	z,Fsctrn	;not our disk, jump to FBIOS
	;
	ld	l,c		;L=C
	ld	h,b		;H=B (no translation)
	ret			;and we're done
	;
;*************************************************************
;Function:  read
;Purpose:   Read logical CP/M sector
;Entry:     disk, track, sector, DMA, etc. set up
;Exit:      sector transferred to memory
;Uses:      All
;
hread:	call	chkhrd		;is it our disk?
	jp	z,Fread		;no, jump to Floppy
	;
	LD	A,wrual		;treat as write to unallocated
	LD	(wrtype),A	;store at WRTYPE
	LD	(READOP),A	;indicate a read (READOP)
alloc0:	XOR	A		;0 in A
	LD	(UNACNT),A	;clear unacnt
alloc1:	INC	A		;bump to 1
	;
noovf1:	LD	(RSFLAG),A	;flag a rd(!0) or wr(0) (rsflag)
;
; calculate host sector...
;
rwoper:	ld	a,(SecShf)	;get sector shift
	ld	b,a		;into B
	or	a		;check for 0
	ld	hl,(SEKSEC)	;get sector
	jr	z,rwope1	;jrif 0 (don't shift)
rwope0:	or	a		;clear carry
	rr	h		;shift sector right H:0->carry
	rr	l		;shift sector right carry->L:7
	djnz	rwope0		;until found
rwope1:	ld	(SEKHST),HL	;store host sector to seek
	XOR	A		;0 in A
	LD	(erflag),A	;no errors (yet)
	LD	HL,HSTACT	;point at hstact
	OR	(HL)		;OR in the host active flag
	LD	(HL),1		;always becomes 1
	JR	Z,filhst	;JRIF was not already 1 (filhst)
	;
	LD	HL,SEKHST	;HL points to SEKHST
	LD	DE,HSTSEC	;DE points HSTSEC
	LD	BC,5		;length to move/compare in BC
	call	strcmp		;compare
	JR	Z,match		;jrif same buffer, disk, track
;nomatch:
	LD	A,(HSTWRT)	;match, host written?
	OR	A		;
	CALL	NZ,wrhst	;flush pending write
	;
filhst:	LD	HL,SEKHST	;point at SEKHST
	LD	DE,HSTSEC	;point to HSTSEC stuff
	LD	BC,5		;count of data to move
	LDIR			;set up HSTDSK
	LD	A,(RSFLAG)	;GET RSFLAG
	OR	A		;Need to read?
	CALL	NZ,rdhst	;rdhst
	XOR	A		;0 in A
	LD	(HSTWRT),A	;no pending write
	;
match:	LD	HL,(SecMsk)	;get mask value in L
	LD	A,(SEKSEC)	;get SEKSEC
	AND	L		;mask with L (A has sector offset)
	LD	BC,00080H	;sector size in BC
	LD	HL,hdbuf	;buffer base in HL?
	RRA			;A /= 2 (transfer A:0 to Carry)
	LD	D,A		;in D  (make high byte of offset)
	LD	E,B		;0 in E
	RR	E		;E /= 2	(rotate Carry into E:7)
	ADD	HL,DE		;make address of CP/M sector in HL
	LD	DE,(DMAADR)	;get DMA address in DE
	LD	A,(READOP)	;get READOP
	OR	A		;is this a READ?
	JR	NZ,rwmove	;yes, skip
	EX	DE,HL		;no, swap pointers
	INC	A		;
	LD	(HSTWRT),A	;hstwrt = 1
rwmove:	LDIR			;move data to/from host buffer
	LD	A,(wrtype)	;get WRTYPE
	DEC	A		;bump down
	LD	A,(erflag)	;get erflag?
	RET	NZ		;return if WRTYPE was wrall or wrual
	OR	A		;was wrdir, errors?
	RET	NZ		; return if errors
	LD	(HSTWRT),A	;no errors, mark buffer written
	jp 	wrhst	;and go to wrhst sector...
	;
;*************************************************************
;Function:  write
;Purpose:   Write logical CP/M sector
;Entry:     disk, track, sector, DMA, etc. set up
;Exit:      sector transferred to disk..sort of
;Uses:      All
;
hwrite:	call	chkhrd		;is it our disk?
	jp	z,Fwrite	;no, jump to Floppy
	;
	XOR	A		;0 in A
	LD	(READOP),A	;indicate a write operation (READOP)
	LD	A,C		;get the write type in A
	LD	(wrtype),A	;store at WRTYP
	CP	wrual		;is it write to unallocated?
	JR	NZ,chkuna	;no, jump over (to chkuna)
	;
	LD	HL,SEKDSK	;point to SEKDSK
	LD	DE,UNADSK	;set pointer to UNADSK
	LD	BC,005H		;count of data to move
	LDIR			;copy SEKDSK --> UNADSK
	LD	A,(HDBLM)	;get BLM	
	INC	A		;bump up by 1
	LD	(UNACNT),A	;set up unacnt
chkuna:	LD	HL,UNACNT	;point at unacnt
	XOR	A		;0 in A
	CP	(HL)		;any more unallocated sectors?
	JP	Z,alloc1	;JRIF no more (alloc1)
	DEC	(HL)		;--unacnt
	LD	HL,SEKDSK	;HL points to SEKDSK
	LD	DE,UNADSK	;point to unadsk disk data
	LD	BC,005H		;count of data to move/compare
	CALL	strcmp		;compare with SEKDSK
	JP	NZ,alloc0	;JRIF not same disk (alloc0)
	;
	LD	HL,(UNASEC)	;get unasec
	INC	HL		;bump up 1 
	LD	(UNASEC),HL	;put it back
	LD	DE,(DPBHD)	;get cp/m sec/trk
	or	a		;clear carry
	sbc	hl,de		;check for overflow
	JR	C,noovf0	;JRIF no overflow 
	;
	ld	hl,0000		;overflow, 0 in HL
	LD	(UNASEC),HL	;reset unasec to 0
	LD	HL,(UNATRK)	;get UNATRK
	INC	HL		;bump to next track
	LD	(UNATRK),HL	;and put it back
noovf0:	XOR	A		;0 in A
	JP	noovf1		; and do a write
	;
;*************************************************************
;Function:  wrhst
;Purpose:   Do physical write to IDE drive
;Entry:     hstdsk, hsttrk, hstdsk set up
;Exit:      sector in hdbuf written to disk
;Used:      all
;
wrhst:	call	tfsetup		;set up Task File
	ld	a,CmdWr		;get a write command
	out	(IdeCmd),a	;send it to Task File
	call	wtrdy		;wait for data
	otir
	otir			;write 512 bytes
	jr	chkerr		;check for/report errors
	;
;*************************************************************
;Function:  rdhst
;Purpose:   Do physical read of IDE drive
;Entry:     hstdsk, hsttrk, hstdsk set up
;Exit:      sector read into hdbuf
;Used:      all
;
rdhst:	call	tfsetup		;set up Task File
rdhs1:	ld	a,CmdRd		;get a read command
	out	(IdeCmd),a	;send to task file
	call	wtrdy		;wait for data
	inir
	inir			;read 512 bytes
chkerr:	
	in	a,(IdeStat)	;get status
	and	1		;check error bit
	ld	(erflag),a	;save as error flag
	ret			;return error/no error
	;
;*************************************************************
;Function:  TFsetup
;Purpose:   send host parms to IDE Task File registers
;Entry:     hstdsk, hsttrk, hstsec set up
;Exit:      Task file registers set up.  HL points to hdbuf,
;	    BC has count, data port address
;
tfsetup:
	in	a,(IdeStat)	;get status
	rla			;BSY bit to Carry
	jr	c,tfsetup	;loop until not BUSY
	;
	ld	a,1		;1-sector transfer
	out	(IdeSCnt),a	;tell IDE drive
	ld	a,(hstsec)	;get host sector number
	inc	a		;bump up to 1
	out	(IdeSNum),a	;send to Task File
	ld	hl,(hsttrk)	;get host track
	ld	a,l		;get low order byte
	out	(IdeCLo),a	;send to Task File
	ld	a,h		;get high-order byte
	out	(IdeCHi),a	;send to task file
	ld	a,(hstdsk)	;get host disk number
	ld	c,a		;put in c
	call	mkesel		;get the select data in A
	out	(IdeSDH),a	;send to drive!
	;
	ld	hl,hdbuf	;HL points to host buffer
	ld	bc,IdeDat	;BC has count, port
	ret			;and go back
	;	
;*************************************************************
;Function:  chkhrd
;Purpose:   check if disk operation is for hard disk or floppy
;Entry:     "ourdisk" is set for hard disk, clear for floppy
;Exit:      Z set if floppy operation, clear if hard disk
;Uses:      AF
;
chkhrd:	ld	a,(hdisk)	;is it our disk?
	or	a		;set flagz
	ret			;and done
	;
;*************************************************************
;Function:  wtrdy
;Purpose:   Wait until data available or ready for data
;Entry:     None
;Exit:      None
;Uses:      AF
;
Wtrdy:	in	a,(IdeCmd)
	bit	3,a
	jr	z,wtrdy
	ret
	;
;*************************************************************
;Function:  strcmp
;Purpose:   Compare two strings [(SEK|UNA|HST)TRK]
;Entry:     HL, DE have addresses of strings, BC = length
;Exit:	    Z set if strings are the same, clear if different
;		P/V flag clear if entire string checked
;Uses:	    AF, BC, DE, HL
;
strcmp:
	LD	A,(DE)		;get a byte from (DE)
	CPI			;compare A-(HL++), BC--
	RET	NZ		;return if no match
	RET	PO		;return if string exhausted
	INC	DE		;else bump to next source byte
	JR	strcmp		;and repeat comparison
	;
;*************************************************************
;
hdisk:	defs	1	;flag if disk operation is for us.
;
; DO NOT CHANGE THE ORDER OF THESE DATA AREAS!!!
;
hstsec:	defs	2		;host sector
hstdsk:	defs	1		;host disk
hsttrk:	defs	2		;host track
sekhst:	defs	2		;host sector to seek
sekdsk:	defs	1		;disk to seek
sektrk:	defs	2		;track to seek
seksec:	defs	2		;sector to seek
unadsk:	defs	1		;unallocated disk
unatrk:	defs	2		;unallocated track
unasec:	defs	2		;unallocated sector
unacnt:	defs	1		;unallocated sector count
hstact:	defs	1		;host active flag
hstwrt:	defs	1		;host written flag
erflag:	defs	1		;error reporting flag
rsflag:	defs	1		;read sector flag
readop:	defb	0
wrtype:	defb	0
dmaadr:	defw	0
;
;-------------------------------------------------------------
; DPH und DPB GIDE
;-------------------------------------------------------------
	;
	;DPH fuer IDE-Laufwerke
	;
	defm 'DPH und DPB'
	;
HDPBASE:
HDPH0:	defw	0000		;XLT0
	defw	0000
	defw	0000, 0000	;scratch
	defw	DIRBF
	defw	DPBHD0
	defw	CSV0
	defw	ALV0
	;
HDPH1:	defw	0000		;XLT1
	defw	0000
	defw	0000, 0000	;scratch
	defw	DIRBF
	defw	DPBHD
	defw	CSV1
	defw	ALV1
	;
HDPH2:	defw	0000		;XLT2
	defw	0000
	defw	0000, 0000	;scratch
	defw	DIRBF
	defw	DPBHD
	defw	CSV2
	defw	ALV2
	;
HDPH3:	defw	0000		;XLT3
	defw	0000
	defw	0000, 0000	;scratch
	defw	DIRBF
	defw	DPBHD
	defw	CSV3
	defw	ALV3
	;
HDPH4:	defw	0000		;XLT4
	defw	0000
	defw	0000, 0000	;scratch
	defw	DIRBF
	defw	DPBHD
	defw	CSV4
	defw	ALV4
	;
HDPH5:	defw	0000		;XLT5
	defw	0000
	defw	0000, 0000	;scratch
	defw	DIRBF
	defw	DPBHD
	defw	CSV5
	defw	ALV5
	;
HDPH6:	defw	0000		;XLT6
	defw	0000
	defw	0000, 0000	;scratch
	defw	DIRBF
	defw	DPBHD
	defw	CSV6
	defw	ALV6
	;
HDPH7:	defw	0000		;XLT7
	defw	0000
	defw	0000, 0000	;scratch
	defw	DIRBF
	defw	DPBHD
	defw	CSV7
	defw	ALV7
	;
	;
	;Host Parameters for 64MB IDE drive...
	;
SecMsk:	defb	3		;512/128 - 1
Secshf:	defb	2		;log_2(512/128)
	;
	;Disk Parameter Block fuer 8MB Laufwerke mit 8 kB Bloecken
	;Erster DPB fuer erstes LW mit 2 Systemspuren
	;
DPBHD0:	defw	128	;SPT 2	;32x4 CP/M Sectors/track (128 Byte)
	defb	6	;BSH 1	;Block Shift Factor (Anz. 128 Byte je Block, 2_log)
	defb	63	;BLM 1	;Block Mask
	defb	3	;EXM 1	;Extent Mask
	defw	3E3h	;DSM 2	;996 blocks
	defw	3FFh	;DRM 2	;1024 directory entries
	defw	0FFh	;ALL 2	;8 blocks of directory !!!
	defw	0	;CKS 2	;Check Area Size (0 for HDD)
	defw	2	;OFF 2	;2 tracks offset (2 Systemspuren)
	;
	;Zweiter DPB fuer weitere LW ohne Systemspuren
	;
DPBHD:	defw	128	;SPT 2	;32x4 CP/M Sectors/track (128 Byte)
	defb	6	;BSH 1	;Block Shift Factor (Anz. 128 Byte je Block, 2_log)
HDBLM:	defb	63	;BLM 1	;Block Mask
	defb	3	;EXM 1	;Extent Mask
	defw	3E7h	;DSM 2	;1000 blocks	(500 Zylinder/Sektoren)
	defw	3FFh	;DRM 2	;1024 directory entries
	defw	0FFh	;ALL 2	;8 blocks of directory !!!
	defw	0	;CKS 2	;Check Area Size (0 for HDD)
	defw	0	;OFF 2	;2 tracks offset
	;
	defm 'CSV und ALV 0-7'
	;
CSV0:	defb 	0
ALV0:	defs	128
	;
CSV1:	defb 	0
ALV1:	defs	128
	;
CSV2:	defb 	0
ALV2:	defs	128
	;
CSV3:	defb 	0
ALV3:	defs	128
	;
CSV4:	defb 	0
ALV4:	defs	128
	;
CSV5:	defb 	0
ALV5:	defs	128
	;
CSV6:	defb 	0
ALV6:	defs	128
	;
CSV7:	defb 	0
ALV7:	defs	128
	;
	;
	defm 'DIRBUF 128 Byte'
	;
dirbf:	defs	128		;Directorypuffer, 128 Byte
	;
	defm 'HDBUF 512 Byte'
	;
hdbuf:	defs 	512		;Lesepuffer HDD
	;
LAST:	defb 	0	
	;
	defm 'ENDE PUFFER'
	;
;	END
	;
	;
	;
